/**
 *  Copyright (C) 2013 Emerson Ribeiro de Mello e Luiz Gustavo Ender
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright (C) 2013 Emerson Ribeiro de Mello e Luiz Gustavo Ender
 *
 *  Este programa é software livre; você pode redistribuí-lo e/ou
 *  modificá-lo sob os termos da Licença Pública Geral GNU, conforme
 *  publicada pela Free Software Foundation; tanto a versão 2 da
 *  Licença como (a seu critério) qualquer versão mais nova.
 *
 *  Este programa é distribuído na expectativa de ser útil, mas SEM
 *  QUALQUER GARANTIA; sem mesmo a garantia implícita de
 *  COMERCIALIZAÇÃO ou de ADEQUAÇÃO A QUALQUER PROPÓSITO EM
 *  PARTICULAR. Consulte a Licença Pública Geral GNU para obter mais
 *  detalhes.
 *
 *  Você deve ter recebido uma cópia da Licença Pública Geral GNU
 *  junto com este programa; se não, escreva para a Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307, USA.
*/

package frame;

import elements.ClassFul;
import elements.Filter;
import elements.Objeto;
import elements.Classe;
import elements.ClassLess;
import java.awt.Dimension;
import java.awt.Point;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import javax.swing.JTextArea;
import org.netbeans.api.visual.widget.ComponentWidget;
import org.netbeans.api.visual.widget.Widget;
import org.openide.util.Exceptions;
import org.openide.util.ImageUtilities;
import org.openide.xml.XMLUtil;
import org.w3c.dom.*;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * Classe responsável por salvar e ler um projeto.
 * 
 * @author Luiz Gustavo Ender
 */
public class VisualSerializer {

    private static final String SCENE_ELEMENT = "Scene"; // NOI18N
    private static final String VERSION_ATTR = "version"; // NOI18N
    private static final String SCENE_NODE_COUNTER_ATTR = "nodeIDcounter"; // NOI18N
    private static final String SCENE_EDGE_COUNTER_ATTR = "edgeIDcounter"; // NOI18N
    private static final String PROJECT_NAME = "ProjectName";
    private static final String PROJECT_IFACE = "ProjectIface";
    private static final String NODE_ELEMENT = "Node"; // NOI18N
    private static final String FILTER_ELEMENT = "Filter";
    private static final String NOTE_ELEMENT = "Note";
    private static final String NOTE_WIDTH = "Width";
    private static final String NOTE_HEIGHT = "Height";
    private static final String NOTE_TEXT = "Text";
    private static final String NODE_ID_ATTR = "id"; // NOI18N
    private static final String NODE_EDGE_ATTR = "Edge";
    private static final String NODE_X_ATTR = "x"; // NOI18N
    private static final String NODE_Y_ATTR = "y"; // NOI18N
    private static final String NODE_TIPO_ATTR = "Tipo";
    private static final String NODE_NIVEL_ATTR = "Nivel";
    private static final String EDGE_ELEMENT = "Edge"; // NOI18N
    private static final String EDGE_ID_ATTR = "id"; // NOI18N
    private static final String EDGE_SOURCE_ATTR = "source"; // NOI18N
    private static final String EDGE_TARGET_ATTR = "target"; // NOI18N
    private static final String VERSION_VALUE_1 = "1"; // NOI18N
    private static final String CONF_ELEMENT = "Conf";
    private static VisualScene scene;
    private static String text;

    /**
     * Função que salva um projeto.
     * @param scene ambiente de trabalho
     * @param file arquivo que onde será salvo
     * @throws Exception caso houver erro.
     */
    public static void serialize(VisualScene scene, File file) throws Exception {
        Document document = XMLUtil.createDocument(SCENE_ELEMENT, null, null, null);
        Node sceneElement = document.getFirstChild();
        setAttribute(document, sceneElement, VERSION_ATTR, VERSION_VALUE_1);
        setAttribute(document, sceneElement, PROJECT_NAME, scene.getProjeto());
        setAttribute(document, sceneElement, PROJECT_IFACE, scene.getIface());
        setAttribute(document, sceneElement, SCENE_NODE_COUNTER_ATTR, Long.toString(scene.getNodeIDcounter()));
        setAttribute(document, sceneElement, SCENE_EDGE_COUNTER_ATTR, Long.toString(scene.getEdgeIDcounter()));

        ArrayList<Objeto> ordenaNode = new ArrayList<Objeto>();
        for (ObjetoNode node : scene.getNodes()) {
            ordenaNode.add((Objeto) scene.findWidget(node));
        }

        for (Objeto widget : ordenaNode) {
            ObjetoNode node = (ObjetoNode) scene.findObject(widget);
            Element nodeElement = document.createElement(NODE_ELEMENT);
            setAttribute(document, nodeElement, NODE_ID_ATTR, node.getNomImagem());
            setAttribute(document, nodeElement, NODE_EDGE_ATTR, node.getEdge());
            Point location = widget.getPreferredLocation();
            setAttribute(document, nodeElement, NODE_X_ATTR, Integer.toString(location.x));
            setAttribute(document, nodeElement, NODE_Y_ATTR, Integer.toString(location.y));
            setAttribute(document, nodeElement, NODE_TIPO_ATTR, ((Objeto) widget).getClass().getSimpleName());
            setAttribute(document, nodeElement, NODE_NIVEL_ATTR, "" + ((Objeto) widget).getNivel());
            nodeElement.appendChild(setAttrObjetos(document, document.createElement(CONF_ELEMENT), (Objeto) widget));
            sceneElement.appendChild(nodeElement);

        }

        for (Filter f : scene.getFiltros()) {
            Element filterElement = document.createElement(FILTER_ELEMENT);
            setAttribute(document, filterElement, "Proto", f.getProto());
            setAttribute(document, filterElement, "Dport", "" + f.getDport());
            setAttribute(document, filterElement, "dIP", f.getdIP());
            setAttribute(document, filterElement, "Sport", "" + f.getSport());
            setAttribute(document, filterElement, "sIP", f.getsIP());
            setAttribute(document, filterElement, "Dest", f.getDestString());
            sceneElement.appendChild(filterElement);
        }

        ArrayList<String> ordenaEdge = new ArrayList<String>();
        for (Objeto o : ordenaNode) {
            for (String edge : scene.getEdges()) {
                Objeto s = (Objeto) scene.findWidget(scene.getEdgeSource(edge));
                if (o.equals(s)) {
                    ordenaEdge.add(edge);
                }
            }
        }

        for (String edge : ordenaEdge) {
            Element edgeElement = document.createElement(EDGE_ELEMENT);
            setAttribute(document, edgeElement, EDGE_ID_ATTR, edge);
            ObjetoNode sourceNode = scene.getEdgeSource(edge);
            if (sourceNode != null) {
                setAttribute(document, edgeElement, EDGE_SOURCE_ATTR, sourceNode.getEdge());
            }
            ObjetoNode targetNode = scene.getEdgeTarget(edge);
            if (targetNode != null) {
                setAttribute(document, edgeElement, EDGE_TARGET_ATTR, targetNode.getEdge());
            }
            sceneElement.appendChild(edgeElement);
        }

        for (ComponentWidget lw : scene.getNotas()) {
            Element noteElement = document.createElement(NOTE_ELEMENT);
            JTextArea j = (JTextArea) lw.getComponent();
            String text = j.getText();
            setAttribute(document, noteElement, NOTE_TEXT, text);
            Point location = lw.getPreferredLocation();
            setAttribute(document, noteElement, NODE_X_ATTR, Integer.toString(location.x));
            setAttribute(document, noteElement, NODE_Y_ATTR, Integer.toString(location.y));
            setAttribute(document, noteElement, NOTE_WIDTH, Integer.toString(j.getWidth() + 12));
            setAttribute(document, noteElement, NOTE_HEIGHT, Integer.toString(j.getHeight() + 12));
            sceneElement.appendChild(noteElement);
        }
        FileOutputStream fos = null;
        try {
            fos = new FileOutputStream(file);
            XMLUtil.write(document, fos, "UTF-8"); // NOI18N
        } catch (Exception e) {
            throw e;
        } finally {
            try {
                if (fos != null) {
                    fos.close();
                }
            } catch (Exception e) {
                Exceptions.printStackTrace(e);
            }
        }
    }
    
    /**
     * Função que ler um projeto.
     * 
     * @param scene ambiente onde será reconstruído o projeto.
     * @param file arquivo do projeto.
     * @throws Exception caso houver erro.
     */
    public static void deserialize(VisualScene scene, File file) throws Exception {
        Node sceneElement = getRootNode(file);
        if (!VERSION_VALUE_1.equals(getAttributeValue(sceneElement, VERSION_ATTR))) {
            throw new Exception();
        }

        try {
            scene.setProjeto(getAttributeValue(sceneElement, PROJECT_NAME));
            scene.setIface(getAttributeValue(sceneElement, PROJECT_IFACE));
            scene.setNodeIDcounter(Integer.parseInt(getAttributeValue(sceneElement, SCENE_NODE_COUNTER_ATTR)));
            scene.setEdgeIDcounter(Integer.parseInt(getAttributeValue(sceneElement, SCENE_EDGE_COUNTER_ATTR)));
            scene.getFrame().clearFiltros();
            for (Node element : getChildNode(sceneElement)) {
                if (NODE_ELEMENT.equals(element.getNodeName())) {
                    String node = getAttributeValue(element, NODE_ID_ATTR);
                    int x = Integer.parseInt(getAttributeValue(element, NODE_X_ATTR));
                    int y = Integer.parseInt(getAttributeValue(element, NODE_Y_ATTR));
                    ObjetoNode o = new ObjetoNode(node);
                    o.setEdge(getAttributeValue(element, NODE_EDGE_ATTR));
                    o.setTipo(getAttributeValue(element, NODE_TIPO_ATTR));
                    scene.setNovo(false);
                    Widget nodeWidget = scene.addNode(o);
                    nodeWidget.setPreferredLocation(new Point(x, y));
                    Objeto objeto = setConfObjetos((Objeto) nodeWidget, element.getChildNodes().item(1));
                    objeto.setNivel(Integer.parseInt(getAttributeValue(element, NODE_NIVEL_ATTR)));
                    objeto.setImage(scene.insertTextImage(ImageUtilities.loadImage("/imagens/" + node), text));
                } else if (FILTER_ELEMENT.equals(element.getNodeName())) {
                    scene.getFrame().addFilter(getAttributeValue(element, "Proto"),
                            getAttributeValue(element, "Dport"),
                            getAttributeValue(element, "dIP"),
                            getAttributeValue(element, "Sport"),
                            getAttributeValue(element, "sIP"),
                            getAttributeValue(element, "Dest"));
                } else if (EDGE_ELEMENT.equals(element.getNodeName())) {
                    String edge = getAttributeValue(element, EDGE_ID_ATTR);
                    String sourceNode = getAttributeValue(element, EDGE_SOURCE_ATTR);
                    String targetNode = getAttributeValue(element, EDGE_TARGET_ATTR);

                    Objeto oSource = null;
                    Objeto oTarget = null;
                    ObjetoNode source = null;
                    ObjetoNode target = null;

                    for (ObjetoNode o : scene.getNodes()) {
                        if (sourceNode.equals(o.getEdge())) {
                            source = o;
                            oSource = (Objeto) scene.findWidget(o);
                        } else if (targetNode.equals(o.getEdge())) {
                            target = o;
                            oTarget = (Objeto) scene.findWidget(o);;
                        }
                    }

                    if (oTarget instanceof ClassFul) {
                        ((ClassFul) oTarget).setParent(oSource);
                        if (oSource instanceof Classe) {
                            ((Classe) oSource).setQdisc((ClassFul) oTarget);
                        }
                    } else if (oTarget instanceof ClassLess) {
                        ((ClassLess) oTarget).setParent(oSource);
                        if (oSource instanceof Classe) {
                            ((Classe) oSource).setQdisc((ClassLess) oTarget);
                        }
                    } else if (oTarget instanceof Classe) {
                        ((Classe) oTarget).setParent(oSource);
                        if (oSource instanceof ClassFul) {
                            ((ClassFul) oSource).addClasse((Classe) oTarget);
                        } else if (oSource instanceof Classe) {
                            ClassFul cl = scene.getClassFul(oSource);
                            if (cl != null) {
                                cl.addClasse((Classe) oTarget);
                            }
                        }
                    }
                    oSource.getChild().add(oTarget);
                    source.setEdge(sourceNode);
                    target.setEdge(targetNode);
                    scene.addEdge(edge);
                    scene.setEdgeSource(edge, source);
                    scene.setEdgeTarget(edge, target);
                } else if (NOTE_ELEMENT.equals(element.getNodeName())) {
                    int x = Integer.parseInt(getAttributeValue(element, NODE_X_ATTR));
                    int y = Integer.parseInt(getAttributeValue(element, NODE_Y_ATTR));
                    int w = Integer.parseInt(getAttributeValue(element, NOTE_WIDTH));
                    int h = Integer.parseInt(getAttributeValue(element, NOTE_HEIGHT));
                    scene.createNote(getAttributeValue(element, NOTE_TEXT), new Point(x, y), new Dimension(w, h));
                }
            }
            scene.repaint();
            scene.validate();

            // Atualiza o default para os ClassFul do tipo HTB
            for (ObjetoNode on : scene.getNodes()) {
                Objeto o = (Objeto) scene.findWidget(on);
                if (o instanceof ClassFul) {
                    if (!"".equals(((ClassFul) o).getCdefaultString())) {
                        ((ClassFul) o).atualizaCdefault();
                    }
                }
            }
        } catch (Exception e) {
            throw e;
        }
    }

    /**
     * Função que seta um atributo do xml.
     * 
     * @param xml documento xml.
     * @param node node do xml.
     * @param name name do atributo.
     * @param value valor do atributo;
     */
    private static void setAttribute(Document xml, Node node, String name, String value) {
        NamedNodeMap map = node.getAttributes();
        Attr attribute = xml.createAttribute(name);
        attribute.setValue(value);
        map.setNamedItem(attribute);
    }

    /**
     * Retorna o Node Raiz de um determinado projeto.
     * @param file arquivo do projeto.
     * @return node raiz.
     */
    private static Node getRootNode(File file) {
        FileInputStream is = null;
        try {
            is = new FileInputStream(file);
            Document doc = XMLUtil.parse(new InputSource(is), false, false, new ErrorHandler() {
                public void error(SAXParseException e) throws SAXException {
                    throw new SAXException(e);
                }

                public void fatalError(SAXParseException e) throws SAXException {
                    throw new SAXException(e);
                }

                public void warning(SAXParseException e) {
                    Exceptions.printStackTrace(e);
                }
            }, null);
            return doc.getFirstChild();
        } catch (Exception e) {
            Exceptions.printStackTrace(e);
        } finally {
            try {
                if (is != null) {
                    is.close();
                }
            } catch (IOException e) {
                Exceptions.printStackTrace(e);
            }
        }
        return null;
    }

    /**
     * Retorna um valor de um documento xml.
     * 
     * @param node node que está o atributo.
     * @param attr nome do atributo.
     * @return retorna o valor do atributo.
     * @throws Exception caso houver erro de leitura,
     */
    private static String getAttributeValue(Node node, String attr) throws Exception {
        try {
            if (node != null) {
                NamedNodeMap map = node.getAttributes();
                if (map != null) {
                    node = map.getNamedItem(attr);
                    if (node != null) {
                        return node.getNodeValue();
                    } else {
                        throw new Exception();
                    }
                }
            }
        } catch (Exception ex) {
            throw ex;
        }
        return null;
    }

    /**
     * Retorna os nodes "filhos" de um node raiz.
     * 
     * @param node node raiz.
     * @return Retorna os nodes "filhos" de um node raiz.
     */
    private static Node[] getChildNode(Node node) {
        NodeList childNodes = node.getChildNodes();
        Node[] nodes = new Node[childNodes != null ? childNodes.getLength() : 0];
        for (int i = 0; i < nodes.length; i++) {
            nodes[i] = childNodes.item(i);
        }
        return nodes;
    }

    /**
     * Utilizado com for salvar um arquivo.
     * Seta os atributos de todos os componentes de QoS
     * 
     * @param xml documento xml.
     * @param e elemento.
     * @param o objeto que contém o conteúdo.
     * @return elemento.
     */
    public static Element setAttrObjetos(Document xml, Element e, Objeto o) {
        if (o instanceof ClassFul) {
            ClassFul cf = (ClassFul) o;
            setAttribute(xml, e, "Handle", cf.getHandle());
            setAttribute(xml, e, "Qdisc", cf.getQdisc());
            if (cf.getImagemNome().equals("qdisc_sub.png")) {
                setAttribute(xml, e, "Parent", cf.getParentString());
            }
            if ("htb".equals(cf.getQdisc())) {
                setAttribute(xml, e, "Default", cf.getCdefaultString());
            }
        } else if (o instanceof ClassLess) {
            ClassLess cl = (ClassLess) o;
            setAttribute(xml, e, "Handle", cl.getHandle());
            setAttribute(xml, e, "Qdisc", cl.getQdisc());
            if (cl.getImagemNome().equals("qdisc_sub.png")) {
                setAttribute(xml, e, "Parent", cl.getParentString());
                if ("pfifo".equals(cl.getQdisc())) {
                    setAttribute(xml, e, "Limit", "" + cl.getLimit());
                } else if ("sfq".equals(cl.getQdisc())) {
                    setAttribute(xml, e, "Quantum", "" + cl.getQuantum());
                    setAttribute(xml, e, "Pertub", "" + cl.getPertub());
                } else {
                    setAttribute(xml, e, "Rate", "" + cl.getRate());
                    setAttribute(xml, e, "Burst", "" + cl.getBurst());
                    setAttribute(xml, e, "Latency", "" + cl.getLatency());
                    setAttribute(xml, e, "Peakrate", "" + cl.getPeakrate());
                    setAttribute(xml, e, "Mtu", "" + cl.getMtu());
                }
            }
        } else if (o instanceof Classe) {
            Classe c = (Classe) o;
            if (c.getNomClass().equals("prio")) {
                setAttribute(xml, e, "Classid", c.getClassid());
                setAttribute(xml, e, "NomClass", c.getNomClass());
                setAttribute(xml, e, "Band", "" + c.getBand());
            } else {
                setAttribute(xml, e, "Parent", c.getParentString());
                setAttribute(xml, e, "Classid", c.getClassid());
                setAttribute(xml, e, "NomClass", c.getNomClass());
                setAttribute(xml, e, "Rate", "" + c.getRate());
                setAttribute(xml, e, "Ceil", "" + c.getCeil());
                if (c.getImagemNome().equals("class_sub.png")) {
                    setAttribute(xml, e, "Burst", "" + c.getBurst());
                    setAttribute(xml, e, "CBurst", "" + c.getCburst());
                    setAttribute(xml, e, "Prio", "" + c.getPrio());
                }
            }
        }
        return e;
    }

    /**
     * Utilizado quando carregado um projeto. Restaura as configurações dos objetos.
     * 
     * @param o objeto em questão.
     * @param n node em questão.
     * @return Retorna o objeto com os parâmetros setados.
     * @throws Exception caso houver erro de leitura.
     */
    public static Objeto setConfObjetos(Objeto o, Node n) throws Exception {
        if (o instanceof ClassFul) {
            ClassFul cf = (ClassFul) o;
            cf.setHandle(getAttributeValue(n, "Handle"));
            cf.setLabel(getAttributeValue(n, "Handle"));
            cf.setQdisc(getAttributeValue(n, "Qdisc"));
            if ("htb".equals(cf.getQdisc())) {
                cf.setCdefault(getAttributeValue(n, "Default"));
            }
            cf.setToolTipText("Qdisc " + cf.getQdisc().toUpperCase());
            text = "Qdisc\n" + cf.getQdisc().toUpperCase();
            return cf;
        } else if (o instanceof ClassLess) {
            ClassLess cl = (ClassLess) o;
            cl.setHandle(getAttributeValue(n, "Handle"));
            cl.setLabel(getAttributeValue(n, "Handle"));
            cl.setQdisc(getAttributeValue(n, "Qdisc"));
            if (cl.getImagemNome().equals("qdisc_sub.png")) {
                cl.setParent(getAttributeValue(n, "Parent"));
                if ("pfifo".equals(cl.getQdisc())) {
                    cl.setLimit(Integer.parseInt(getAttributeValue(n, "Limit")));
                } else if ("sfq".equals(cl.getQdisc())) {
                    cl.setQuantum(Integer.parseInt(getAttributeValue(n, "Quantum")));
                    cl.setPertub(Integer.parseInt(getAttributeValue(n, "Pertub")));
                } else {
                    cl.setRate(Integer.parseInt(getAttributeValue(n, "Rate")));
                    cl.setBurst(Integer.parseInt(getAttributeValue(n, "Burst")));
                    cl.setLatency(Integer.parseInt(getAttributeValue(n, "Latency")));
                    cl.setPeakrate(Integer.parseInt(getAttributeValue(n, "Peakrate")));
                    cl.setMtu(Integer.parseInt(getAttributeValue(n, "Mtu")));
                }
            }
            cl.setToolTipText("Qdisc " + cl.getQdisc().toUpperCase());
            text = "Qdisc\n" + cl.getQdisc().toUpperCase();
            return cl;
        } else if (o instanceof Classe) {
            Classe c = (Classe) o;
            if (getAttributeValue(n, "NomClass").equals("prio")) {
                c.setLabel(getAttributeValue(n, "Classid"));
                c.setClassid(getAttributeValue(n, "Classid"));
                c.setNomClass(getAttributeValue(n, "NomClass"));
                c.setBand(Integer.parseInt(getAttributeValue(n, "Band")));
            } else {
                c.setParent(getAttributeValue(n, "Parent"));
                c.setLabel(getAttributeValue(n, "Classid"));
                c.setClassid(getAttributeValue(n, "Classid"));
                c.setNomClass(getAttributeValue(n, "NomClass"));
                c.setRate(Integer.parseInt(getAttributeValue(n, "Rate")));
                c.setCeil(Integer.parseInt(getAttributeValue(n, "Ceil")));
                if (c.getImagemNome().equals("class_sub.png")) {
                    c.setBurst(Integer.parseInt(getAttributeValue(n, "Burst")));
                    c.setCburst(Integer.parseInt(getAttributeValue(n, "CBurst")));
                    c.setPrio(Integer.parseInt(getAttributeValue(n, "Prio")));
                }
            }
            c.setToolTipText("Classe " + c.getNomClass().toUpperCase());
            text = "Classe\n" + c.getNomClass().toUpperCase();
            return c;
        }
        return null;
    }
}
